/*****************************************************************
play & record audio

$Id: plax.c,v 1.14 1999/02/23 12:44:05 purnhage Exp $

Source file: plax.c

(c) Heiko Purnhagen 1998
e-mail: purnhage@tnt.uni-hannover.de
WWW:    http://www.tnt.uni-hannover.de/~purnhage/

All usual disclaimers apply!

This program may be copied according to the GNU General Public Licence.
See file COPYING for further details.

Authors:
HP    Heiko Purnhagen, Uni Hannover <purnhage@tnt.uni-hannover.de>

Changes:
16-sep-98   HP      first version
18-sep-98   HP      mono/stereo and sample rate conversion
21-sep-98   HP      fixed numSample%2=0 for stereo rate conversion
30-dec-98   HP      CVS, evaluates USE_AFSP
07-jan-99   HP      AFsp-v4r1
08-jan-99   HP      audio device selection
11-jan-99   HP      startTime
17-jan-99   HP      ...
22-jan-99   HP      "audio device -> audio device" error message
26-jan-99   HP      improved output file format evaluation
28-jan-99   HP      fixed bufLen
18-feb-99   HP      progress in seconds / "plax" -> reset audio device
21-feb-99   HP      -d 1 gives "plax: * -> *" message
23-feb-99   HP      fixed debug output ...
**********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>

#ifdef USE_AFSP
#include <libtsp.h>		/* AFsp audio file library */
#include <libtsp/AFpar.h>	/* AFsp audio file library - definitions */
#endif

#include "common.h"		/* common module */
#include "cmdline.h"		/* command line module */
#include "austream.h"		/* audio i/o streams */
#include "playrec.h"		/* audio play & record */
#include "rateconv.h"		/* audio sample rate conversion */


/* ---------- declarations ---------- */

#define PROGVER "plax V0.5.3 23-Feb-1999"
#define CVSID "$Id: plax.c,v 1.14 1999/02/23 12:44:05 purnhage Exp $"
#ifndef DISTVER		/* set in Makefile */
#define DISTVER "<unknown>"
#endif

#define min(a,b) ((a) < (b) ? (a) : (b))
#define max(a,b) ((a) > (b) ? (a) : (b))

#define STRLEN (255)

#define FS_MAX (48000.0)
#define BUF_TIME (1.)		/* length of buffer in seconds */

#define RAWENV "RAWAUDIOFILE"
#define RAWDEFAULT "integer16,0,48000.,native,1,1.0"

#ifdef USE_AFSP
#ifdef FW_SUN
/* only AFsp-V3R2 available: map AFsp-V3R2 to AFsp-v4r1 */
#define AFsetNHpar AFsetNH
#define FTW_AU (FW_SUN/256)
#define FTW_WAVE (FW_WAVE/256)
#define FTW_AIFF_C (FW_AIFF_C/256)
#define FTW_NH_EB (FW_NH_EB/256)
#endif
#endif


/* ---------- variables ---------- */

/* command line module */

static char *progName = "<no program name>";
static char *outFileName;
static int outFileNameUsed;
static int outNumChan;
static int outNumChanUsed;
static float outFSample;
static int outFSampleUsed;
static int recNumChan;
static int recNumChanUsed;
static float recFSample;
static int recFSampleUsed;
static char *playDevice;
static int playDeviceUsed;
static char *recDevice;
static int recDeviceUsed;
static int autoConvert;
static int convert;
static float gain;
static float startTime;
static float duration;
static char *rawFormat;
static int rawFormatUsed;
static float bufTime;
static int ht;
static float al;
static float fc;
static float fd;
static int debugLevel;
static int *varArgIdx;

static CmdLinePara paraList[] = {
  {&varArgIdx,NULL,"<input files (- = stdin) (dflt: audio device)>"},
  {NULL,NULL,NULL}
};

static CmdLineSwitch switchList[] = {
  {"h",NULL,NULL,NULL,NULL,"print help"},
  {"o",&outFileName,"%s",NULL,&outFileNameUsed,
   "output file name (- = stdout) (dflt: audio device)"},
  {"c",&outNumChan,"%d",NULL,&outNumChanUsed,
   "output: number of channels (dflt: as input or 1)"},
  {"s",&outFSample,"%f",NULL,&outFSampleUsed,
   "output: sampling frequency [Hz] (dflt: as input or 8000.0)"},
  {"rc",&recNumChan,"%d",NULL,&recNumChanUsed,
   "rec: number of channels\n"
   "(SGI: 0 = use audio panel settings) (dflt: as output)"},
  {"rs",&recFSample,"%f",NULL,&recFSampleUsed,
   "rec: sampling frequency [Hz]\n"
   "(SGI: 0 = use audio panel settings) (dflt: as output)"},
  {"pd",&playDevice,"%s",NULL,&playDeviceUsed,
   "play: audio device name (dflt: default device)"},
  {"rd",&recDevice,"%s",NULL,&recDeviceUsed,
   "rec: audio device name (dflt: default device)"},
  {"a",&autoConvert,NULL,NULL,NULL,
   "enable automatic conversion to audio device capabilities"},
  {"x",&convert,NULL,NULL,NULL,
   "enable channel & sample rate conversion"},
  {"g",&gain,"%f","1.0",NULL,
   "amplitude gain"},
  {"ts",&startTime,"%f","0.0",NULL,
   "start time [sec]"},
  {"td",&duration,"%f","0.0",NULL,
   "duration [sec] (0=unlimited)"},
  {"f",&rawFormat,"%s",NULL,&rawFormatUsed,
   "raw input file format\n"
   "(dflt: $" RAWENV " / \"" RAWDEFAULT "\")"},
  {"b",&bufTime,"%f","0.2",NULL,
   "buffer time [sec] (if no sample rate conv.)"},
  {"ht",&ht,"%d","-1",NULL,
   "sample rate conv.: num taps\n"
   "(-1 = auto)"},
  {"al",&al,"%f","-1.0",NULL,
   "sample rate conv.: alpha for Kaiser window\n"
   "(-1 = auto)"},
  {"fc",&fc,"%f","-1.0",NULL,
   "sample rate conv.: 6 dB cutoff freq / input bandwidth\n"
   "(-1 = auto)"},
  {"fd",&fd,"%f","-1.0",NULL,
   "sample rate conv.: 100 dB cutoff freq / input bandwidth\n"
   "(-1 = auto)"},
  {"d",&debugLevel,"%d","0",NULL,
   "debug level (on stderr) (0=off 1=basic 2=full)\n"
   "(+4*audioDbgLvl +16*playrecDbgLvl +64*rateconvDbgLvl)\n\b"},
  {NULL,NULL,NULL,NULL,NULL,NULL}
};

static int stop = 0;


/* ---------- local functions ---------- */

void stopproc (int sig)
{
  stop = 1;
}


/* ---------- main ---------- */

int main (int argc, char *argv[])
{
  int listIdx;
  char inFileName[STRLEN];
  char fileName[STRLEN];
#ifdef USE_AFSP
  AFILE *inaf,*outaf;
#endif
  AuStream *inas,*outas;
  PRStream *inpr,*outpr;
  RCBuf *rc;
  int result;
  int RCdebugLevel;

  long maxBufLen,bufLen,convLen,bufTrans;
  long inBufUsed,outBufUsed,tmpBufUsed;
  long numClip,totClip;
  short *bufS;
  float *bufF;
  short *outBufS;
  float *outBufF;
  long i,tmp;

  int inNumChan;
  long inNumChanL;
  float inFSample;
  long fileNumSample;
  long inNumSamp;
  long outNumSamp;
  int devNumChan;
  float devFSample;
  long maxNumSamp;
  int fmt;
#ifdef USE_AFSP
  int fmti;
  struct {
    char *str;
    int fmt;
  } fmtstr[] = {
    {".au",FTW_AU*256},
    {".snd",FTW_AU*256},
    {".wav",FTW_WAVE*256},
    {".aif",FTW_AIFF_C*256},
    {".raw",FTW_NH_EB*256},	/* no header big-endian */
    {".AU",FTW_AU*256},
    {".SND",FTW_AU*256},
    {".WAV",FTW_WAVE*256},
    {".AIF",FTW_AIFF_C*256},
    {".RAW",FTW_NH_EB*256},	/* no header big-endian */
    {"au",FTW_AU*256},
    {"snd",FTW_AU*256},
    {"wav",FTW_WAVE*256},
    {"aif",FTW_AIFF_C*256},
    {"raw",FTW_NH_EB*256},	/* no header big-endian */
    {"AU",FTW_AU*256},
    {"SND",FTW_AU*256},
    {"WAV",FTW_WAVE*256},
    {"AIF",FTW_AIFF_C*256},
    {"RAW",FTW_NH_EB*256},	/* no header big-endian */
    {NULL,-1}
  };
#else
  int fmti;
  struct {
    char *str;
    int fmt;
  } fmtstr[] = {
    {".au",1},
    {".snd",1},
    {".AU",1},
    {".SND",1},
    {"au",1},
    {"snd",1},
    {"AU",1},
    {"SND",1},
    {NULL,-1}
  };
#endif
  int actOutNumChan;
  int actConvert;
  float actOutFSample;

  typedef enum {AUDEV, AUSTREAM, AUFILE} autype;
  autype inty,outty;

  /* evaluate command line  */
  CmdLineInit(0);
  result = CmdLineEval(argc,argv,paraList,switchList,1,&progName);
  if (result) {
    if (result==1) {
      fprintf(stderr,"%s: %s\n"
	      "%s\n",
	      progName,PROGVER "   " DISTVER,CVSID);
      CmdLineHelp(progName,paraList,switchList,stderr);
      exit (1);
    }
    else
      CommonExit(1,"command line error (\"-h\" for help)");
  }
  AuInit((debugLevel>>2)&3);
  PRInit((debugLevel>>4)&3);
  RCdebugLevel = (debugLevel>>6)&3;
  debugLevel &= 3;

  if (debugLevel)
    fprintf(stderr,"%s: %s\n",progName,PROGVER "   " DISTVER);
  if (debugLevel>1) {
    fprintf(stderr,"%s\n",CVSID);
    fprintf(stderr,"debugLevel=%d\n",debugLevel);
  }

#ifdef USE_AFSP
  if (rawFormatUsed)
    AFsetNHpar(rawFormat);
  else {
    if (getenv(RAWENV))
      AFsetNHpar("$" RAWENV);
    else
      AFsetNHpar(RAWDEFAULT);
  }
#endif

  if (!recDeviceUsed)
    recDevice = NULL;
  if (!playDeviceUsed)
    playDevice = NULL;

  maxBufLen = (long)(BUF_TIME*FS_MAX*2);
  if(!(bufF = (float *)malloc(maxBufLen*sizeof(float))))
    CommonExit(1,"failed to allocate bufF");
  if(!(bufS = (short *)malloc(maxBufLen*sizeof(short))))
    CommonExit(1,"failed to allocate bufS");
  outBufS = bufS;

  if (!outFileNameUsed && varArgIdx[0]<0) {
    CommonWarning("nothing to play -> just resetting audio device ...");
    devNumChan = outNumChan = (outNumChanUsed)?outNumChan:1;
    devFSample = outFSample = (outFSampleUsed)?outFSample:8000.0;
    if (!(outpr=PROpenWrite(playDevice,&devNumChan,&devFSample)))
      CommonExit(1,"can not open output audio device");
    if (devNumChan != outNumChan || devFSample != outFSample)
      CommonWarning("can not configure output audio device as requested");
    PRClose(outpr);
    exit (0);
  }
  
  if (outFileNameUsed && varArgIdx[0]>=0 && varArgIdx[1]>=0)
    CommonExit(1,"max one input file allowed when using -o");
  
  signal(SIGINT,stopproc);

  /* process all files on command line or record from audio device */
  listIdx = 0;
  while (!stop && (outFileNameUsed || varArgIdx[listIdx] >= 0)) {

    if (debugLevel)
      fprintf(stderr,
	      "%s: %s -> %s\n",progName,
	      (varArgIdx[listIdx]>=0)?argv[varArgIdx[listIdx]]:"audio device",
	      (outFileNameUsed)?outFileName:"audio device");

    /* open input */
    if (varArgIdx[listIdx]>=0) {
      if (ComposeFileName(argv[varArgIdx[listIdx]],
			  0,NULL,NULL,inFileName,STRLEN))
	CommonExit(1,"composed input file name too long \"%s\"",argv[listIdx]);
      if (strcmp(inFileName,"-")==0) {
	inty = AUSTREAM;
	if (!(inas=AuOpenRead(inFileName,&inNumChan,&inFSample,
			      &fileNumSample)))
	  CommonExit(1,"can not open input \"%s\"",inFileName);
      }
      else {
#ifdef USE_AFSP
	inty = AUFILE;
	if (!(inaf=AFopenRead(inFileName,&fileNumSample,&inNumChanL,
			      &inFSample,
			      (debugLevel>1)?stderr:(FILE*)NULL)))
	  CommonExit(1,"can not open input \"%s\"",inFileName);
	inNumChan = inNumChanL;
#else
	inty = AUSTREAM;
	if (!(inas=AuOpenRead(inFileName,&inNumChan,&inFSample,
			      &fileNumSample)))
	  CommonExit(1,"can not open input \"%s\"",inFileName);
#endif
      }
    }
    else {
      inty = AUDEV;
      fileNumSample = -1;
      inNumChan = (outNumChanUsed)?outNumChan:1;
      inFSample = (outFSampleUsed)?outFSample:8000.0;
      if (recNumChanUsed)
	inNumChan = recNumChan;
      if (recFSampleUsed)
	inFSample = recFSample;
      if (!(inpr=PROpenRead(recDevice,&inNumChan,&inFSample)))
	CommonExit(1,"can not open input audio device");
      if (inNumChan<=0 && outNumChanUsed)
	inNumChan = outNumChan;
      if (inFSample<=0 && outFSampleUsed)
	inFSample = outFSample;
      if (inNumChan<=0 || inFSample<=0)
	CommonExit(1,"input audio device configuration troubles");
    }
    if (debugLevel>1)
      fprintf(stderr,
	      "input=%s inNumChan=%d inFSample=%.1f fileNumSample=%ld\n",
	      (inty==AUDEV)?"audio device":inFileName,
	      inNumChan,inFSample,fileNumSample);
    
    /* open output */
    if (!outNumChanUsed)
      outNumChan = inNumChan;
    if (!outFSampleUsed)
      outFSample = inFSample;
    actOutNumChan = outNumChan;
    actOutFSample = outFSample;
    actConvert = convert;

    if (outFileNameUsed) {
      if (ComposeFileName(outFileName,0, NULL,NULL,fileName,STRLEN))
	CommonExit(1,"composed output file name too long \"%s\"",outFileName);
      if (strcmp(fileName,"-")==0) {
	outty = AUSTREAM;
	if (!(outas=AuOpenWrite(fileName,actOutNumChan,actOutFSample)))
	  CommonExit(1,"can not open output \"%s\"",fileName);
      }
      else {
#ifdef USE_AFSP
	outty = AUFILE;
	fmti = 0;
	while (fmtstr[fmti].str && strstr(fileName,fmtstr[fmti].str)==NULL)
	  fmti++;
	if (fmtstr[fmti].str)
	  fmt = FD_INT16 + fmtstr[fmti].fmt;
	else
	  CommonExit(1,"unkown output audio in format");
	if (!(outaf=AFopenWrite(fileName,fmt,actOutNumChan,actOutFSample,
				(debugLevel>1)?stderr:(FILE*)NULL)))
	  CommonExit(1,"can not open output \"%s\"",fileName);
#else
	outty = AUSTREAM;
	fmti = 0;
	while (fmtstr[fmti].fmt!=-1 && strstr(fileName,fmtstr[fmti].str)==NULL)
	  fmti++;
	if (!fmtstr[fmti].str)
	  CommonExit(1,"unkown output audio in format");
	if (!(outas=AuOpenWrite(fileName,actOutNumChan,actOutFSample)))
	  CommonExit(1,"can not open output \"%s\"",fileName);
#endif
      }
    }
    else {
      outty = AUDEV;
      devNumChan = actOutNumChan;
      devFSample = actOutFSample;
      if (!(outpr=PROpenWrite(playDevice,&devNumChan,&devFSample)))
	CommonExit(1,"can not open output audio device");
      if (devNumChan != actOutNumChan || devFSample != actOutFSample) {
	CommonWarning("can not configure output audio device as requested");
	if (autoConvert) {
	  convert = 1;
	  actOutNumChan = devNumChan;
	  actOutFSample = devFSample;
	  CommonWarning("=> converting from %dch/%.0fHz to %dch/%.0fHz",
			outNumChan,outFSample,actOutNumChan,actOutFSample);
	}
	else
	  CommonWarning("=> using %dch/%.0fHz instead of %dch/%.0fHz",
			devNumChan,devFSample,actOutNumChan,actOutFSample);
      }
    }
    if (debugLevel>1)
      fprintf(stderr,"output=%s actOutNumChan=%d actOutFSample=%.1f\n",
	      (outty==AUDEV)?"audio device":fileName,
	      actOutNumChan,actOutFSample);

    /* init buffers & conversion */
    if (convert && (inNumChan > 2 || actOutNumChan > 2))
      CommonExit(-1,"file conversion only for mono/stereo");
    bufLen = (long)(inNumChan*inFSample*bufTime+.5);
    bufLen = max(bufLen,inNumChan);
    bufLen = (bufLen/inNumChan)*inNumChan;
    rc = NULL;
    if (convert && (inFSample!=actOutFSample || fc>=0 || fd>=0)) {
      if (!(rc=RateConvInit(RCdebugLevel,actOutFSample/(float)inFSample,
			    min(inNumChan,actOutNumChan),ht,al,fc,fd,
			    &convLen)))
	CommonExit(1,"can not init sample rate conversion");
      bufLen = convLen;
      if (inNumChan==2 && actOutNumChan==1)
	bufLen *= 2;
    }
    bufLen = min(bufLen,maxBufLen);
    if (debugLevel>2)
      fprintf(stderr,"bufLen=%ld maxBufLen=%ld\n",bufLen,maxBufLen);

    /* skip samples (startTime) */
    inNumSamp = 0;
    if (startTime <= 0)
      startTime = 0;
    else {
      maxNumSamp = (long)(inNumChan*inFSample*startTime+.5);
      do {

	if (debugLevel) {
	  if (debugLevel<=2) {
	    fprintf(stderr,"\r~%7.3f ",inNumSamp/inNumChan/inFSample);
	    fflush(stderr);
	  }
	  else
	    fprintf(stderr,"~%7.3f ",inNumSamp/inNumChan/inFSample);
	}
	
	bufTrans = maxNumSamp-inNumSamp;
	if (bufTrans>bufLen)
	  bufTrans = bufLen;
	
	switch (inty) {
	case AUSTREAM:
	  inBufUsed = AuReadData(inas,bufS,bufTrans);
	  for (i=0; i<inBufUsed; i++)
	    bufF[i] = bufS[i];
	  break;
	case AUFILE:
#ifdef USE_AFSP
	  inBufUsed = AFreadData(inaf,inNumSamp,bufF,bufTrans);
#endif
	  break;
	case AUDEV:
	  inBufUsed = PRReadData(inpr,bufS,bufTrans);
	  for (i=0; i<inBufUsed; i++)
	    bufF[i] = bufS[i];
	  break;
	}
	inNumSamp += inBufUsed;

	if (debugLevel>2)
	    fprintf(stderr,"inBufUsed=%ld\n",inBufUsed);

      } while (!stop && inBufUsed);
    }

    /* copy samples */
    totClip = 0;
    outNumSamp = 0;
    if (duration <= 0)
      maxNumSamp = 0;
    else
      maxNumSamp = (long)(inNumChan*inFSample*(startTime+duration)+.5);
    do {

      if (debugLevel) {
	if (debugLevel<=2) {
	  fprintf(stderr,"\r%8.3f ",inNumSamp/inNumChan/inFSample);
	  fflush(stderr);
	}
	else
	  fprintf(stderr,"%8.3f ",inNumSamp/inNumChan/inFSample);
      }

      bufTrans = (maxNumSamp)?maxNumSamp-inNumSamp:bufLen;
      if (bufTrans>bufLen)
	bufTrans = bufLen;

      switch (inty) {
      case AUSTREAM:
	inBufUsed = AuReadData(inas,bufS,bufTrans);
	for (i=0; i<inBufUsed; i++)
	  bufF[i] = bufS[i];
	break;
      case AUFILE:
#ifdef USE_AFSP
	inBufUsed = AFreadData(inaf,inNumSamp,bufF,bufTrans);
#endif
	break;
      case AUDEV:
	inBufUsed = PRReadData(inpr,bufS,bufTrans);
	for (i=0; i<inBufUsed; i++)
	  bufF[i] = bufS[i];
	break;
      }
      inNumSamp += inBufUsed;

      outBufUsed = inBufUsed;
      outBufF = bufF;
      if (convert) {
	if (inNumChan==2 && actOutNumChan==1) {
	  outBufUsed = inBufUsed/2;
	  for (i=0; i<outBufUsed; i++)
	    bufF[i] = (bufF[2*i]+bufF[2*i+1])/2;
	}
	if (rc) {
	  while (outBufUsed<convLen)
	    bufF[outBufUsed++] = 0;
	  outBufUsed = RateConv(rc,bufF,outBufUsed,&outBufF);
	  tmpBufUsed = (long)(inNumSamp/inNumChan/inFSample*
			      actOutFSample*actOutNumChan)-outNumSamp;
	  if (actOutNumChan==2) {
	    if (inNumChan==1)
	      tmpBufUsed /= 2;
	    else
	      tmpBufUsed = (tmpBufUsed/actOutNumChan)*actOutNumChan;
	  }
	  if (tmpBufUsed<outBufUsed)
	    outBufUsed = tmpBufUsed;
	}
	if (inNumChan==1 && actOutNumChan==2) {
	  outBufUsed *= 2;
	  for (i=outBufUsed-1; i>=0; i--)
	    bufF[i] = outBufF[i/2];
	  outBufF = bufF;
	}
      }
      if (outBufUsed>maxBufLen)
	CommonExit(-1,"output buffer size troubles");
      numClip = 0;
      for (i=0; i<outBufUsed; i++) {
	tmp = ((long)(gain*outBufF[i]+32768.5))-32768;
	if (tmp>32767) {
	  tmp = 32767;
	  numClip++;
	}
	if (tmp<-32768) {
	    tmp = -32768;
	    numClip++;
	}
	outBufS[i] = tmp;
      }
      totClip += numClip;

      if (debugLevel>2)
	  fprintf(stderr,"inBufUsed=%ld outBufUsed=%ld numClip=%ld\n",
		  inBufUsed,outBufUsed,numClip);

      switch (outty) {
      case AUSTREAM:
	AuWriteData(outas,outBufS,outBufUsed);
	break;
      case AUFILE:
#ifdef USE_AFSP
	AFwriteData(outaf,outBufF,outBufUsed);
#endif
	break;
      case AUDEV:
	PRWriteData(outpr,outBufS,outBufUsed);
	break;
      }
      outNumSamp += outBufUsed;

    } while (!stop && inBufUsed);
      
    if (rc)
      RateConvFree(rc);
      
    /* close input */
    switch (inty) {
    case AUSTREAM:
      AuClose(inas);
      break;
    case AUFILE:
#ifdef USE_AFSP
      AFclose(inaf);
#endif
      break;
    case AUDEV:
      PRClose(inpr);
      break;
    }

    /* close output */
    switch (outty) {
    case AUSTREAM:
      AuClose(outas);
      break;
    case AUFILE:
#ifdef USE_AFSP
      AFclose(outaf);
#endif
      break;
    case AUDEV:
      PRClose(outpr);
      break;
    }

    if (debugLevel && debugLevel<=2) {
      fprintf(stderr," \n");
      fflush(stderr);
    }
    if (debugLevel>1)
      fprintf(stderr,
	      "in: %ld samples (%.3f sec)   "
	      "out: %ld samples (%.3f sec)\n",
	      inNumSamp,inNumSamp/(inNumChan*inFSample),
	      outNumSamp,outNumSamp/(actOutNumChan*actOutFSample));
    if (totClip)
      CommonWarning("%ld samples clipped!",totClip);

    if (outFileNameUsed)
      outFileNameUsed = 0;
    if (varArgIdx[listIdx]>=0)
      listIdx++;
  }

  CmdLineEvalFree(paraList);

  if (debugLevel)
    fprintf(stderr,"%s: %s\n",progName,stop?"aborted":"finished");

  exit (0);
}

/* end of plax.c */

